import java.util.Random;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;

 /**
 *  <h4>Canon.java</h4>
 *
 *  Copyright (C) 2015 Guido Kramann<br/>
 *  kramann@fh-brandenburg.de<br/>
 *  http://www.kramann.info<br/>
 *  <br/>
 *  <p>This implementation is optimized for an easy understanding but it is not very efficient.</p>
 *  <br/>
 *  <p>  
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  </p>
 *  <p>
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  </p>
 *  <p>
 *  You should have received a copy of the GNU General Public License
 *  aint with this program (gpl.html).  If not, see http://www.gnu.org/licenses/.
 *  </p>
 */
public class Canon extends UnifyingCounterpointandHarmonics
{
    public static int COUNT = 0;

    public static Random rand = new Random(System.currentTimeMillis());

    public static int midimin = 55;
    public static int midimax = 72;
    public static int maxtonenumber = 1;

    public static void changeMelos(int[] mel)
    {
         int anz = rand.nextInt(4)+1;   
         int hoehe = midimin + rand.nextInt(midimax-midimin+1-4*anz);      
         for(int i=0;i<anz;i++)
         {
             int stelle = rand.nextInt(mel.length);
             hoehe += rand.nextInt(5); //verschiedene Töne nehmen

             mel[stelle]=hoehe*100;

             //microtonal change
             if(rand.nextInt(100)==20)
                 mel[stelle]+=-2+rand.nextInt(5);
         }
         reduceNumberOfTones(mel,maxtonenumber);
    }

    public static int consonance(int[][] mattrans)
    {
        int c=0;
        for(int i=0;i<mattrans.length;i++)
           c+=chordConsonance(mattrans[i]);
        return c;
    }
    public static int restlessness(int[][] mattrans)
    {
        int sum=0;
        int[] c=new int[mattrans[0].length];
        int[] d=new int[mattrans[0].length];
        for(int ii=0;ii<=mattrans.length;ii++)
        {
           int i = ii%mattrans.length;
           for(int k=0;k<mattrans[i].length;k++)
               d[k] = tonePersistenceMicro(k,mattrans[i]);

           if(ii>0)
           {
               for(int k=0;k<mattrans[i].length;k++)
                   sum+=abs(d[k]-c[k]);
           }

           for(int k=0;k<c.length;k++)
               c[k]=d[k];
        }
        return sum;
    }

    

    public static int selfsimilarity(int[] vek)
    {
        int[] raster = {6*2,12*2,18*2,24*2};
        int gesamt=0;
        for(int r=0;r<raster.length;r++)
        {
            int best=0;
            for(int phase=0;phase<raster[r];phase++)
            {
                int ae=0;
                for(int p=0;p<vek.length;p+=raster[r]) //Anfang
                {
                    int gleich=0;                  
                    int ungleich=0;                  
                    for(int q=0;q<raster[r]/2;q++)
                    {
                        int qa = (p+q)%vek.length;
                        int qb = (p+q+raster[r]/2)%vek.length;
                        int qa2 = (qa+1)%vek.length;
                        int qb2 = (qb+1)%vek.length;
                        
                        if(vek[qa]==vek[qb])
                            gleich++;
                        if(vek[qa]!=vek[qa2])
                            ungleich++;
                        if(vek[qb]!=vek[qb2])
                            ungleich++;
                    }
                    if(ungleich>1)
                        ae+=(gleich*2)*(ungleich);
                }
                if(ae>best)
                    best=ae;
            }
            gesamt+=best;
        }
        return gesamt;
    }

    public static void main(String[] args)
    {
        //initial conditions

        int[] melos           = new int[96];
        int[] diffmelos       = new int[96];
        int[][] matrix          = new int[4][melos.length];
        int[][] matrixtrans     = new int[melos.length][4];

        int[] testmelos       = new int[melos.length];
        int[] testdiffmelos   = new int[melos.length];
        int[][] testmatrix      = new int[matrix.length][melos.length];
        int[][] testmatrixtrans = new int[melos.length][matrix.length];

        int[] displacement    = new int[] {24,0,12,36};
        int[] transposition   = new int[] {700,0,-500,-1200};

        maxtonenumber = (melos.length*2)/5;
        for(int i=0;i<melos.length;i++)
            melos[i] = 100*(midimin + rand.nextInt(midimax-midimin+1));
        copyMelosToMatrix(melos,matrix,displacement,transposition);
        reduceNumberOfTones(melos,maxtonenumber);
        transpose(matrix,matrixtrans);
        differentiate(melos,diffmelos);
        int cons = consonance(matrixtrans);
        int rest = restlessness(matrixtrans);
        int self1= selfsimilarity(melos);
        int self2= selfsimilarity(diffmelos);
        save("canon"+COUNT+"_c"+cons+"_r"+rest+"_sa"+self1+"_sb"+self2+".txt",matrix);                   
        //optimization
        while(true)
        {
            //new variant
            copy(melos,testmelos);
            changeMelos(testmelos);
            copyMelosToMatrix(testmelos,testmatrix,displacement,transposition);
            transpose(testmatrix,testmatrixtrans);
            differentiate(testmelos,testdiffmelos);
            //examine it
            int cons_new   = consonance(testmatrixtrans);
            int rest_new   = 0;   
            int self1_new   = 0; 
            int self2_new  = 0;        
            boolean OPT = false;
            if(cons_new >= cons)
            {
                OPT=true;
                rest_new   = restlessness(testmatrixtrans);
                self1_new  = selfsimilarity(testmelos);
                self2_new  = selfsimilarity(testdiffmelos);
            }
            else if(cons_new >= cons-10)
            {
                rest_new   = restlessness(testmatrixtrans);
                if(rest_new >= rest)
                {
                    OPT=true;
                    self1_new  = selfsimilarity(testmelos);
                    self2_new  = selfsimilarity(testdiffmelos);
                }
                else if(rest_new >= rest-10)
                {
                    self1_new  = selfsimilarity(testmelos);
                    if(self1_new >= self1)    
                    {
                        OPT=true;
                        self2_new  = selfsimilarity(testdiffmelos);
                    }
                    else if(self1_new >= self1-10)
                    {
                        self2_new  = selfsimilarity(testdiffmelos);
                        if(self2_new >= self2)
                        {
                            OPT=true;
                        }
                    }
                }
            }

            if(OPT==true)
            {
               copy(testmelos,melos);
               copy(testmatrix,matrix);
               cons = cons_new;
               rest = rest_new;
               self1= self1_new;
               self2= self2_new; 

               COUNT++;
               if(COUNT%10==0)
               {
                   System.out.println("cons="+cons+" rest="+rest+" self1="+self1+" self2="+self2);
                   save("canon"+COUNT+"_c"+cons+"_r"+rest+"_sa"+self1+"_sb"+self2+".txt",matrix);                   
               }
            }
        }
    }

    //**** Helper Methods ****

    public static void copy(int[][] source, int[][] sink)
    {
        for(int i=0;i<source.length;i++)
            for(int k=0;k<source[i].length;k++)
                sink[i][k] = source[i][k];
    }
    public static void copy(int[] source, int[] sink)
    {
        for(int i=0;i<source.length;i++)
                sink[i] = source[i];
    }

    public static void copyMelosToMatrix(int[] mel, int[][] mat, int[] disp, int[] trans)
    {
          for(int i=0;i<mat.length;i++)
                for(int k=0;k<mat[i].length;k++)
                    mat[i][(k+disp[i])%mat[i].length]=mel[k]+trans[i];
    }
    public static int countTones(int[] mel)
    {
         int anz=0;
         for(int i=0;i<mel.length;i++)
         {
             if(mel[i]>0 && mel[(i-1+mel.length)%mel.length]!=mel[i])
                 anz++;
         }
         return anz;    
    }
    public static void reduceNumberOfTones(int[] mel, int max)
    {
        while(countTones(mel)>max)
        {
             int i = rand.nextInt(mel.length);
             int k = (i+1+mel.length)%mel.length;
             mel[k]=mel[i];
        }
    }

    public static void transpose(int[][] mat, int[][] mattrans)
    {
        for(int i=0;i<mat.length;i++)
            for(int k=0;k<mat[i].length;k++)
                mattrans[k][i]=mat[i][k];
    }

    public static int abs(int x)
    {
        if(x<0)
            return -x;
        else
            return x;
    }

    public static void differentiate(int[] mel, int[] diffmel)
    {
        for(int i=0;i<mel.length;i++)
            diffmel[i]=mel[i]-mel[(i-1+mel.length)%mel.length];
    }

    public static void save(String name,int[][] mat)
    {
        try
        {
            File              meinFile = new File(name);
            FileWriter   meinSchreiber = new FileWriter(meinFile);
            PrintWriter  meinPrinter   = new PrintWriter( meinSchreiber );            
            for(int i=0;i<mat.length;i++)
            {
                for(int k=0;k<mat[i].length;k++)
                {
                    meinPrinter.print(mat[i][k]);            
                    meinPrinter.print(" ");            
                }
                meinPrinter.println();            
            }
            meinPrinter.close();   
        }
        catch(IOException e)
        {
            System.out.println(e.getMessage());
        }

    }
}