import java.util.ArrayList;

 /**
 *  <h4>UnifyingCounterpointandHarmonics.java</h4>
 *
 *  Copyright (C) 2015 Guido Kramann<br/>
 *  kramann@fh-brandenburg.de<br/>
 *  http://www.kramann.info<br/>
 *  <br/>
 *  <p>This implementation is optimized for an easy understanding but it is not very efficient.</p>
 *  <br/>
 *  <p>  
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  </p>
 *  <p>
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  </p>
 *  <p>
 *  You should have received a copy of the GNU General Public License
 *  aint with this program (gpl.html).  If not, see http://www.gnu.org/licenses/.
 *  </p>
 */
public class UnifyingCounterpointandHarmonics
{
    //CHAPTER 1

    /**midi2frequency(mp) = (440/(2^(69/12)))*2^(mp/12)*/
    public static double midi2Frequency(double mp)
    {
        double freq69 = Math.pow(2.0,69.0/12.0);
        return Math.pow(2.0,mp/12.0)*440.0/freq69;
    }

    /**frequency2midi(f) = 12*ld( f * ((2^(69/12))/440) )*/
    public static double frequency2Midi(double f)
    {
        double freq69 = Math.pow(2.0,69.0/12.0);
        return 12.0*Math.log(f*freq69/440.0)/Math.log(2.0);
    }

    /**midicent2frequency(mcp) = (440/(2^(69/12)))*2^(mcp/1200)*/
    public static double midicent2Frequency(double mcp)
    {
        double freq69 = Math.pow(2.0,69.0/12.0);
        return Math.pow(2.0,mcp/1200.0)*440.0/freq69;
    }

    /**frequency2midicent(f) = 1200*ld( f * ((2^(69/12))/440) )*/
    public static double frequency2Midicent(double f)
    {
        double freq69 = Math.pow(2.0,69.0/12.0);
        return 1200.0*Math.log(f*freq69/440.0)/Math.log(2.0);
    }

    /**midi2midicent(mp) = 100 * mp*/
    public static double midi2midicent(double mp)
    {
        return 100.0*mp;
    }

    /**midicent2midi(mcp) = mcp/100*/
    public static double midicent2midi(double mcp)
    {
        return mcp/100.0;
    }

    //CHAPTER 3

    /**
    *   dock[ ] = dockingPoints(mcp,N)                 <br/>
    *     f=midicent2frequency(mcp)                    <br/>
    *     for i=0:N-1                                  <br/>
    *         dock[i]=frequency2midicent(f/(N+1-i))    <br/>
    *     end                                          <br/>
    *     dock[N]=mcp                                  <br/>
    *     for i=0:N-1                                  <br/>
    *         dock[i+N+1]=frequency2midicent(f*(i+2))  <br/>
    *     end                                          <br/>
    *     return dock[ ]                               <br/>
    *   end.                                           <br/>
    */
    public static int[] dockingPoints(int mcp, int N) 
    {
        int[] dock = new int[N*2+1];
        dock[N]=mcp;
        double f = midicent2Frequency((double)mcp);
        for(int i=0;i<N;i++)
        {
            double fo = (double)(i+2)*f;
            double fu = f/(double)(i+2);
            double mco = frequency2Midicent(fo);
            double mcu = frequency2Midicent(fu);
            
            dock[N-i-1]=(int)Math.round(mcu);            
            dock[N+i+1]=(int)Math.round(mco);            
        }
        return dock;
    }

    //CHAPTER 4

    /** combine <br/>
    *   Combines two arrays and sorts the resulting array.
    *   It also takes care that each value in the resulting array only appears once.
    */
    public static int[] combine(int[] a, int[] b)
    {
        ArrayList<int[]> liste = new ArrayList<int[]>();
        for(int i=0;i<a.length;i++)
        {
            boolean gefunden=false;
            for(int k=0;k<liste.size();k++)
            {
                 if(liste.get(k)[0]==a[i])
                 {
                     gefunden=true;
                     break;
                 }
            }
            if(gefunden==false)
            {
                liste.add(new int[] {a[i]});
            }
        }
        for(int i=0;i<b.length;i++)
        {
            boolean gefunden=false;
            for(int k=0;k<liste.size();k++)
            {
                 if(liste.get(k)[0]==b[i])
                 {
                     gefunden=true;
                     break;
                 }
            }
            if(gefunden==false)
            {
                liste.add(new int[] {b[i]});
            }
        }
        
        int[] erg = new int[liste.size()];
        for(int i=0;i<erg.length;i++)
            erg[i] = liste.get(i)[0];
        liste.clear();    
        //Sortieren
        
        boolean gefunden=true;
        while(gefunden==true)
        {
            gefunden=false;
            for(int i=0;i<erg.length-1;i++)
            {
                if(erg[i]>erg[i+1])
                {
                    gefunden=true;
                    int h=erg[i];
                    erg[i]=erg[i+1];
                    erg[i+1]=h;
                }
            }
        }
        
        return erg;
    }


    /**
    *   dockmult[ ] = dockingPointsMult(mcp,sref[ ])                               <br/>
    *     N=(size(sref[ ])-1)/2                                                    <br/>
    *     dockmult[ ] = dockingPoints(mcp,N)                                       <br/>
    *     mcpOct=mcp+1200                                                          <br/>
    *     while frequency2Midicent(midicent2Frequency(mcpOct)/N)%lt;=max(sref[ ]) do  <br/>
    *         dockmult[ ] = combine( dockmult[ ] , dockingPoints(mcpOct,N) )       <br/>
    *         mcpOct=mcpOct+1200                                                   <br/>
    *     end                                                                      <br/>
    *     mcpOct=mcp-1200                                                          <br/>
    *     while frequency2Midicent(midicent2Frequency(mcpOct)*N)&gt;=min(sref[ ]) do  <br/>
    *         dockmult[ ] = combine( dockmult[ ] , dockingPoints(mcpOct,N) )       <br/>
    *         mcpOct=mcpOct-1200                                                   <br/>
    *     end                                                                      <br/>
    *     return dockmult[ ]                                                       <br/>
    *   end.                                                                       <br/>
    */
    public static int[] dockingPointsMult(int midicent, int[] referenz)
    {
        int max = referenz[referenz.length-1];
        int min = referenz[0];
        int anz = (referenz.length-1)/2;

        int[] erg = dockingPoints(midicent,anz);
        
        int midicent_H = midicent+1200;
        int midicent_HX = (int)frequency2Midicent(midicent2Frequency((double)midicent_H)/(double)anz);
        while(midicent_HX<=max)
        {
            int[] h = dockingPoints(midicent_H,anz);
            erg = combine(erg,h);
            midicent_H+=1200;
            midicent_HX = (int)frequency2Midicent(midicent2Frequency((double)midicent_H)/(double)anz);
        }
        int midicent_L = midicent-1200;
        int midicent_LX = (int)frequency2Midicent(midicent2Frequency((double)midicent_L)*(double)anz);
        while(midicent_LX>=min)
        {
            int[] l = dockingPoints(midicent_L,anz);
            erg = combine(erg,l);
            midicent_L-=1200;
            midicent_LX = (int)frequency2Midicent(midicent2Frequency((double)midicent_L)*(double)anz);
        }
        return erg;
    }

    /**
    *   count = tonalCoincidence(mcp,mcpref,TOLERANCE=24,N=5)                          <br/>
    *     count=0                                                                      <br/>
    *     sref[ ]=dockingPoints(mcpref,N)                                              <br/>
    *     s=dockingPointsMult(mcp,sref[ ])                                             <br/>
    *     foreach s[ ] as v                                                            <br/>
    *         foreach sref[ ] as vref                                                  <br/>
    *             if v+TOLERANCE&gt;=vref AND v-TOLERANCE&lt;=vref THEN count=count+1  <br/>
    *         end                                                                      <br/>
    *     end                                                                          <br/>
    *     return count                                                                 <br/>
    *   end.                                                                           <br/>
    */
    public static int tonalCoincidence(int mcp, int mcpref)
    {
        int TOLERANCE = 24;
        int N = 5;
        return tonalCoincidence(mcp,mcpref,TOLERANCE,N);
    }
    public static int tonalCoincidence(int mcp, int mcpref, int TOLERANCE, int N)
    {
        int count=0;
        int[] sref = dockingPoints(mcpref,N);
        int[] s    = dockingPointsMult(mcp,sref);
        for(int i=0;i<s.length;i++)
            for(int k=0;k<sref.length;k++)
                if(s[i]+TOLERANCE>=sref[k] && s[i]-TOLERANCE<=sref[k])
                    count++;
        return count;            
    }

    //CHAPTER 5

    /**
    *   count = tonalSymmetricCoincidence(mcpA,mcpB,TOLERANCE=24,N=5)                              <br/>
    *     return tonalCoincidence(mcpA,mcpB,TOLERANCE,N) + tonalCoincidence(mcpB,mcpA,TOLERANCE,N) <br/>
    *   end.                                                                                       <br/>
    */
    public static int tonalSymmetricCoincidence(int mcpA, int mcpB)
    {
        int TOLERANCE = 24;
        int N = 5;
        return tonalSymmetricCoincidence(mcpA,mcpB,TOLERANCE,N);
    }
    public static  int tonalSymmetricCoincidence(int mcpA, int mcpB, int TOLERANCE, int N)
    {
        return tonalCoincidence(mcpA,mcpB,TOLERANCE,N) + tonalCoincidence(mcpB,mcpA,TOLERANCE,N);            
    }

    /**
    *   count = tonalSymmetricCoincidenceMicro(mcpA,mcpB,TOLERANCE=24,N=5)             <br/>
    *     c = tonalSymmetricCoincidence(mcpA,mcpB,TOLERANCE,N)                         <br/>
    *     micro = 0                                                                    <br/>
    *     TOLERANCE = TOLERANCE - 1                                                    <br/>
    *     while TOLERANCE &gt;= 0 AND c == tonalSymmetricCoincidence(mcpA,mcpB,TOLERANCE,N) <br/>
    *         micro = micro + 1                                                        <br/>
    *         TOLERANCE = TOLERANCE - 1                                                <br/>
    *     end                                                                          <br/>
    *     return c*100 + micro                                                         <br/>
    *   end.                                                                           <br/>
    */
    public static int tonalSymmetricCoincidenceMicro(int mcpA,int mcpB)
    {
        int TOLERANCE = 24;
        int N = 5;
        return tonalSymmetricCoincidenceMicro(mcpA,mcpB,TOLERANCE,N);
    }
    public static int tonalSymmetricCoincidenceMicro(int mcpA,int mcpB,int TOLERANCE, int N)
    {
        int c = tonalSymmetricCoincidence(mcpA,mcpB,TOLERANCE,N);
        int micro = 0;
        TOLERANCE--;
        while(TOLERANCE>=0 && c == tonalSymmetricCoincidence(mcpA,mcpB,TOLERANCE,N))
        {
            micro++;
            TOLERANCE--;
        }
        return c*100 + micro;
    }


    /**
    *   count = chordConsonance(chord[ ],TOLERANCE=24,N=5)                             <br/>
    *     c = 0                                                                        <br/>
    *     m = size(chord[ ])                                                           <br/>
    *     for i=0:m-1                                                                  <br/>
    *         minimum=tonalSymmetricCoincidenceMicro(chord[i],chord[i],TOLERANCE,N)    <br/>
    *         for k=0:m-1                                                              <br/>
    *             test = tonalSymmetricCoincidenceMicro(chord[i],chord[k],TOLERANCE,N) <br/>
    *             if i!=k AND test&lt;=minimum                                              <br/>
    *                 minimum =  test                                                  <br/>
    *             end                                                                  <br/>
    *         end                                                                      <br/>
    *         c = c + minimum                                                          <br/>
    *     end                                                                          <br/>
    *     return c/m                                                                   <br/>
    *   end.                                                                           <br/>
    */
    public static int chordConsonance(int[] chord)
    {
        int TOLERANCE = 24;
        int N = 5;
        return chordConsonance(chord,TOLERANCE,N);
    }
    public static int chordConsonance(int[] chord,int TOLERANCE,int N)
    {
        int c = 0;
        for(int i=0;i<chord.length;i++)
        {
            int minimum = tonalSymmetricCoincidenceMicro(chord[i],chord[i],TOLERANCE,N);
            for(int k=0;k<chord.length;k++)
            {
                int test = tonalSymmetricCoincidenceMicro(chord[i],chord[k],TOLERANCE,N);
                if(i!=k && test<=minimum)
                    minimum = test;
            }
            c+=minimum;
        }
        return c/chord.length;
    }

    public static int betrag(int x)
    {
        if(x<0)
            return -x;
        else
            return x;
    }

    public static void sort(int[] x)
    {
        boolean gefunden = true;
        while(gefunden)
        {
            gefunden = false;
            for(int i=0;i<x.length-1;i++)
            {
                if(x[i]>x[i+1])
                {
                    int h = x[i];
                    x[i]=x[i+1];
                    x[i+1]=h;
                    gefunden=true;
                }
            }
        }
    }


    /**
    *   count = tonePersistence(index, chord[ ],TOLERANCE=24,N=5)               <br/>
    *     count = 0                                                             <br/>
    *     m = size(chord[ ])                                                    <br/>
    *     q = 0                                                                 <br/>
    *     testtone = chord[index]                                               <br/>
    *     for i=0:m-1                                                           <br/>
    *         if i!=index                                                       <br/>
    *             if q==0                                                       <br/>
    *                 dockref[ ] = dockingPoints(chord[i],N)                    <br/>
    *             else                                                          <br/>
    *                 dockref[ ] = combine( dock , dockingPoints(chord[i],N) )  <br/>
    *             end                                                           <br/>
    *             q=q+1                                                         <br/>
    *         end                                                               <br/>
    *     end                                                                   <br/>
    *     dock = dockingPointsMult(testtone,dockref[ ])                         <br/>
    *     foreach dock[ ] as v                                                  <br/>
    *         foreach dockref[ ] as vref                                        <br/>
    *             if v+TOLERANCE&gt;=vref AND v-TOLERANCE&lt;=vref THEN count=count+1 <br/>
    *         end                                                               <br/>
    *     end                                                                   <br/>
    *     return count                                                          <br/>
    *   end.                                                                    <br/>
    */
    public static int tonePersistence(int index,int[] chord)
    {
        int TOLERANCE = 24;
        int N = 5;
        return tonePersistence(index,chord,TOLERANCE,N);
    }
    public static int tonePersistence(int index,int[] chord, int TOLERANCE, int N)
    {
        int count = 0;
        int m     = chord.length;
        int q     = 0;
        int testtone = chord[index];
        int[] dockref = new int[] {};
        for(int i=0;i<chord.length;i++)
        {
            if(i!=index)
            {
                if(q==0)
                    dockref = dockingPoints(chord[i],N);
                else
                    dockref = combine( dockref , dockingPoints(chord[i],N) );
                q++;
            }
        }
        int[] dock = dockingPointsMult(testtone,dockref);
        for(int i=0;i<dock.length;i++)
            for(int k=0;k<dockref.length;k++)
                if(dock[i]+TOLERANCE>=dockref[k] && dock[i]-TOLERANCE<=dockref[k])
                    count++;
        return count;
    }

    /**
    *   count = tonePersistenceMicro(index, chord[ ],TOLERANCE=24,N=5)                   <br/>
    *     count = tonePersistence(index, chord[ ],TOLERANCE,N)                           <br/>
    *     micro = 0                                                                      <br/>
    *     TOLERANCE = TOLERANCE - 1                                                      <br/>
    *     while TOLERANCE &gt;= 0 AND count == tonePersistence(index, chord[ ],TOLERANCE,N) <br/>
    *         micro = micro + 1                                                          <br/>
    *         TOLERANCE = TOLERANCE - 1                                                  <br/>
    *     end                                                                            <br/>
    *     return count*100 + micro                                                       <br/>
    *   end.                                                                             <br/>
    */
    public static int tonePersistenceMicro(int index,int[] chord)
    {
        int TOLERANCE = 24;
        int N = 5;
        return tonePersistenceMicro(index,chord,TOLERANCE,N);
    }
    public static int tonePersistenceMicro(int index,int[] chord,int TOLERANCE,int N)
    {
         int count = tonePersistence(index,chord,TOLERANCE,N);
         int micro = 0;
         TOLERANCE--;
         while(TOLERANCE >= 0 && count == tonePersistence(index, chord,TOLERANCE,N))
         {
             micro++;
             TOLERANCE--; 
         }
         return count*100 + micro;
    }

    /**
    *    some tests
    */ 
    public static void main(String[] args)
    {
        System.out.println("**** Basic Methods ****");

        System.out.println("midi2Frequency(69)="+midi2Frequency(69));

        System.out.println("frequency2Midi(440.0)="+frequency2Midi(440.0));

        System.out.println("midicent2Frequency(6900)="+midicent2Frequency(6900));

        System.out.println("frequency2Midicent(440.0)="+frequency2Midicent(440.0));

        System.out.println("midi2midicent(69)="+midi2midicent(69));

        System.out.println("midicent2midi(6900)="+midicent2midi(6900));    

        System.out.println("**** Analysing Intervals ****");

        int[] dock = dockingPoints(6900,5); 
        showArray("Docking points for mcp=6900mc",dock);
        int[] dockMult = dockingPointsMult(6900+700,dock); 
        showArray("Multiple docking points for mcp=6900+700mc with 6900mc as reference",dockMult);

        System.out.println("tonalCoincidence(6900+700,6900)="+tonalCoincidence(6900+700,6900));

        for(int i=0;i<=24;i++)
        {
            System.out.println("tonalSymmetricCoincidence(6000,"+(6000+i*100)+")="+tonalSymmetricCoincidence(6000,6000+i*100));    
        }

        int[][] chords = {
                             {6000,7200},  //unisono with 2
                             {6000,7200,84000},  //unisono with 3
                             {6000,6700},  //fith
                             {6000,6700},  //fith
                             {6000,6702},  //better fith
                             {6000,6400},  //major third
                             {6000,6300},  //minor third
                             {6000,6400,6700}, //C-Major
                             {6000,6400,5500}, //C5-Major
                             {6000,5200,5500}, //C3-Major
                             {6000,6300,6700}, //C-Minor
                             {6000,6500,6700}, //CFG
                             {6000,6500,6900}, //CFA
                             {6000,6400,6700,7000}, //C7
                             {6000,6400,6700,7200}, //C-Major                             
                             {6000,6400,6702,7200}, //C-Major with better fifth                            
                         };

        for(int i=0;i<chords.length;i++)
        {
            int[] chord = chords[i];
            System.out.print("chordConsonance([");
            for(int k=0;k<chord.length;k++)
                System.out.print(chord[k]+", ");
            System.out.println("])="+chordConsonance(chord));
        } 

        System.out.println("**** Linear Counterpoint ****");
        
        System.out.println("Persistence of E in C-G:"+tonePersistenceMicro(1,new int[] {6000,6400,6700}));
        System.out.println("Persistence of C in E-G:"+tonePersistenceMicro(0,new int[] {6000,6400,6700}));
        System.out.println("Persistence of G in C-E:"+tonePersistenceMicro(2,new int[] {6000,6400,6700}));
        System.out.println();
        System.out.println("Persistence of F in C-A:"+tonePersistenceMicro(1,new int[] {6000,6500,6900}));
        System.out.println("Persistence of C in F-A:"+tonePersistenceMicro(0,new int[] {6000,6500,6900}));
        System.out.println("Persistence of A in C-F:"+tonePersistenceMicro(2,new int[] {6000,6500,6900}));
        System.out.println();
        System.out.println("Persistence of D in H-F:"+tonePersistenceMicro(1,new int[] {5900,6200,6500}));
        System.out.println("Persistence of H in D-F:"+tonePersistenceMicro(0,new int[] {5900,6200,6500}));
        System.out.println("Persistence of F in H-D:"+tonePersistenceMicro(2,new int[] {5900,6200,6500}));
        System.out.println();
        System.out.println("Persistence of c' in c-c'':"+tonePersistenceMicro(1,new int[] {4800,6000,7200}));
        System.out.println("Persistence of c in c'-c'':"+tonePersistenceMicro(0,new int[] {4800,6000,7200}));
        System.out.println("Persistence of c'' in c-c':"+tonePersistenceMicro(2,new int[] {4800,6000,7200}));
        System.out.println();
    }

    //Helper Methods

    public static void showArray(String name,int[] x)
    {
        System.out.println("\nArray "+name+":");
        for(int i=0;i<x.length;i++)
            System.out.print(x[i]+" ");
        System.out.println();    
    }
}
